/*
 * notary - cryptographically notarizing tool
 * Copyright (C) 2000  Tim Waugh <twaugh@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * The idea is that you run this as an inetd service.  When someone wants
 * a document notarized, they sign it and send the detached signature
 * to this program, which then counter-signs it.
 *
 * The counter-signature contains a monotonically increasing sequence
 * number.
 *
 * Of course, the machine on which this program is running must be
 * trusted.  It would help if the administrator used a daily cron job
 * to put the current sequence number and the date into a file (signed
 * using the same key that the notary uses) which is available by some
 * means (HTTP, FTP, etc.).  This prevents suspicion that the
 * administrator has tampered with the system clock.  Ideally, archived
 * sequence-date files would be available, with the detached signatures
 * of those available by some _other_ means (such as mail or news).
 *
 * I wrote this because I wanted to use the PGP Digital Timestamping
 * service (http://www.itconsult.co.uk/stamper.htm), but didn't
 * understand how to get GnuPG to import their public key (now I know
 * to use --allow-non-selfsigned-uid).
 *
 * If you have the means to, please set this up somewhere so that I
 * can use it. ;-)
 */

#define VERSION "notary 0.0.4"

#include <getopt.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/file.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <syslog.h>
#include <unistd.h>

#define COMMENT "For info see http://people.redhat.com/twaugh/notary/"
#define SEQFILE "/var/spool/notary/seq"
#define SIGFILE "/var/spool/notary/signature.log"
#define MAXDIGITS 8

static const char *comment = COMMENT;

static int teechar (FILE *f, int c)
{
	putchar (c);
	return fputc (c, f);
}

static int teeprintf (FILE *f, char *fmt, ...)
{
	int ret;
	va_list ap;
	va_start (ap, fmt);
	vprintf (fmt, ap);
	ret = vfprintf (f, fmt, ap);
	va_end (ap);
	return ret;
}

static void insert_reference (FILE *sig)
{
	int fd = open (SEQFILE, O_RDWR);
	off_t digits;
	char *f, *p;

	if (fd == -1) {
		syslog (LOG_ERR, "couldn't open \"%s\": %m", SEQFILE);
		return;
	}

	digits = lseek (fd, 0, SEEK_END);
	lseek (fd, 0, SEEK_SET);

	if (digits > MAXDIGITS)
		digits = MAXDIGITS;

	f = mmap (NULL, digits, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
	if (!f) {
		syslog (LOG_ERR, "couldn't mmap \"%s\": %m", SEQFILE);
		return;
	}

	if (lockf (fd, F_LOCK, 0) == -1) {
		syslog (LOG_ERR, "lockf failed: %m");
		return;
	}

	p = f + digits - 1;
	while (*p == '\n' && p > f)
		p--;

	if (p == f) {
		/* All newlines */
		memset (p, '0', digits);
		p += digits - 1;
	}

	while (p >= f) {
		if (*p == '9') {
			/* Carry */
			*p = '0';
			p--;
		} else {
			++*p;
			break;
		}
	}

	teeprintf (sig, "Comment: Notary reference ID: ");
	for (p = f; p < f + digits; p++)
		if (*p != '\n')
			teechar (sig, *p);
	teechar (sig, '\n');
	if (*comment)
		teeprintf (sig, "Comment: %s\n", comment);

	msync (f, digits, MS_SYNC);
	munmap (f, digits);
	lockf (fd, F_ULOCK, 0);
	close (fd);
}

static int notarize (void)
{
	int ret = 0;
	int ch[2];
	int last;
	FILE *fil;
	int fd[2];
	pid_t pid;
	int sigfd;
	FILE *sig;

	sigfd = open (SIGFILE, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
	if (sigfd == -1) {
		syslog (LOG_ERR, "couldn't open \"%s\": %m", SIGFILE);
		return -1;
	}

	if (pipe (fd)) {
		syslog (LOG_ERR, "couldn't create pipe: %m");
		close (sigfd);
		return -1;
	}
	pid = fork ();
	if (pid == -1) {
		syslog (LOG_ERR, "couldn't fork: %m");
		close (fd[0]);
		close (fd[1]);
		close (sigfd);
		return -1;
	}

	if (pid == 0) {
		/* Child. */
		close (sigfd);
		close (fd[0]);

		if (STDOUT_FILENO != fd[1]) {
			close (STDOUT_FILENO);
			dup2 (fd[1], STDOUT_FILENO);
			close (fd[1]);
		}

		execlp ("gpg", "gpg", "--homedir", "/etc/notary", "--batch",
			"--no-secmem-warning", "-ba", NULL);
		syslog (LOG_ERR, "execlp failed: %m");
		exit (1);
	}

	close (fd[1]);
	close (STDIN_FILENO);

	/* Append output to SIGFILE. */
	if (lseek (sigfd, 0, SEEK_SET))
		syslog (LOG_ERR, "lseek failed: %m");

	if (lockf (sigfd, F_LOCK, 0))
		syslog (LOG_ERR, "lockf failed!: %m");

	if (lseek (sigfd, 0, SEEK_END) == (off_t) -1)
		syslog (LOG_ERR, "lseek failed: %m");

	sig = fdopen (sigfd, "r+");
	if (!sig)
		syslog (LOG_ERR, "fdopen failed");

	/* Insert the reference ID in the output. */
	fil = fdopen (fd[0], "rb");
	last = 0;
	ch[0] = fgetc (fil);
	if (ch[0] == EOF) {
		syslog (LOG_ERR, "short output");
		ret = -1;
		goto out;
	}
	teechar (sig, ch[last]);
	do {
		last = 1 - last;
		ch[last] = fgetc (fil);
		if (ch[0] == '\n' && ch[1] == '\n')
			insert_reference (sig);

		if (ch[last] != EOF)
			teechar (sig, ch[last]);
	} while (!feof (fil));

 out:
	fflush (sig);
	lseek (sigfd, 0, SEEK_SET);
	lockf (sigfd, F_ULOCK, 0);
	fclose (sig);
	fclose (fil);
	wait (NULL);
	return ret;
}

int main (int argc, char **argv)
{
	openlog ("notary", 0, LOG_DAEMON);

	for (;;) {
		int option_index;
		static struct option long_options[] =
		{ 
			{"version", 0, 0, 0},
			{"help", 0, 0, 0},
			{"comment", 1, 0, 0},
			{0, 0, 0, 0}
		};
		int c = getopt_long (argc, argv, "",
				     long_options, &option_index);
		if (c == -1)
			break;

		switch (option_index) {
		case 0:
			printf ("%s\n", VERSION);
			exit (0);
		case 1:
			printf ("usage: notary [--comment=comment-string]\n");
			exit (0);
		case 2:
			comment = optarg;
			break;
		}
	}

	notarize ();
	closelog ();
	return 0;
}
